
#import <AddressBook/AddressBook.h>
#import "HKMInviteView.h"
#import "HKMInviteViewCell.h"
#import "HKMInviteViewAccessContactCell.h"
#import "HKMDiscoverer.h"
#import "HKMLead.h"
#import "MBProgressHUD.h"
#import "ODRefreshControl.h"

@interface HKMInviteView (private)
- (void)fadeIn;
- (void)fadeOut;
@end

@implementation HKMInviteView
@synthesize delegate;
BOOL needRefresh = false;
BOOL firstTimeQueryLeadsCompleted = true;

#pragma mark - activate
+ (void) activateWithAppKey:(NSString *)key {
    [HKMDiscoverer activate:key];
}

+ (void) activateWithAppKey:(NSString *)key customParam: (NSString *) name {
    [HKMDiscoverer activateKey:key customParam:name];
}

#pragma mark - initialization & cleaning up
- (id)initWithTitle:(NSString *)aTitle
     sendBtnLabel:(NSString *)sendBtnLabel
			 view:(UIView*)view
    useVirtualNumber:(BOOL)virtualNumber
        inviterName:(NSString *)name
{
    NSLog(@"initWithTitle invoked %@", self);

    if (!virtualNumber && ![HKMDiscoverer deviceSmsSupport])
        virtualNumber = true;

    _useVirtualNumber = virtualNumber;
    _inviterName = [name copy];
    
    CGRect rect = view.bounds;
    if (self = [super initWithFrame:rect])
    {
        self.backgroundColor = [UIColor clearColor];
        _title = [aTitle copy];
        
        // register to receive all AGE notification
        [self registerNotification];
        
        // start discovering address book and request IOS6 address book permission
        [self launchWithPermissionCheckDiscover: YES];
        
        float tableViewWidth = rect.size.width - 2 * POPLISTVIEW_SCREENINSET;
        float tableViewHeight = rect.size.height - 2 * POPLISTVIEW_SCREENINSET - POPLISTVIEW_HEADER_HEIGHT - POPLISTVIEW_FOOTER_HEIGHT- RADIUS;
        
        _tableView = [[UITableView alloc] initWithFrame:CGRectMake(POPLISTVIEW_SCREENINSET,
                                                                   POPLISTVIEW_SCREENINSET + POPLISTVIEW_HEADER_HEIGHT,
                                                                   tableViewWidth,
                                                                   tableViewHeight)];
        _tableView.separatorColor = [UIColor colorWithWhite:0 alpha:.2];
        _tableView.backgroundColor = [UIColor clearColor];
        _tableView.dataSource = self;
        _tableView.delegate = self;
        
        // add pulldown refresh control
        ODRefreshControl *refreshControl = [[ODRefreshControl alloc] initInScrollView:_tableView];
        [refreshControl addTarget:self action:@selector(dropViewDidBeginRefreshing:) forControlEvents:UIControlEventValueChanged];
        
        [self addSubview:_tableView];
        
        // add invite button
        float buttonWidth = (tableViewWidth-POPLISTVIEW_BUTTONINSET*2);
        float buttonHeight = POPLISTVIEW_FOOTER_HEIGHT-POPLISTVIEW_BUTTONINSET*2;
        _sendButton = [UIButton buttonWithType:UIButtonTypeRoundedRect];
        [_sendButton addTarget:self
                        action:@selector(sendInviteAction)
              forControlEvents:UIControlEventTouchUpInside];
        [_sendButton setTitle:[sendBtnLabel copy] forState:UIControlStateNormal];
        _sendButton.frame = CGRectMake(POPLISTVIEW_SCREENINSET+POPLISTVIEW_BUTTONINSET, POPLISTVIEW_SCREENINSET + POPLISTVIEW_HEADER_HEIGHT + tableViewHeight + POPLISTVIEW_BUTTONINSET, buttonWidth, buttonHeight);
        [self addSubview:_sendButton];
    }
    return self;
}

- (id)initWithTitle:(NSString *)aTitle
       sendBtnLabel:(NSString *)sendBtnLabel
               view:(UIView*)view
   useVirtualNumber:(BOOL)virtualNumber
{
    return [self initWithTitle:aTitle sendBtnLabel:sendBtnLabel view:view useVirtualNumber:virtualNumber inviterName:nil];
}

- (void)dropViewDidBeginRefreshing:(ODRefreshControl *)refreshControl
{
    double delayInSeconds = 0.0;
    dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, delayInSeconds * NSEC_PER_SEC);
    dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
        [refreshControl endRefreshing];
    });
    //[MBProgressHUD showHUDAddedTo:self animated:YES];
    //[[HKMDiscoverer agent] queryLeads];
    [self launchWithPermissionCheckDiscover:YES];
}

#pragma mark - Private Methods
- (void)fadeIn
{
    self.transform = CGAffineTransformMakeScale(1.3, 1.3);
    self.alpha = 0;
    [UIView animateWithDuration:.35 animations:^{
        self.alpha = 1;
        self.transform = CGAffineTransformMakeScale(1, 1);
    }];
    if ([HKMDiscoverer agent].installCode == nil)
        [self launchWithPermissionCheckDiscover: YES];
}

- (void)fadeOut
{
    [UIView animateWithDuration:.35 animations:^{
        self.transform = CGAffineTransformMakeScale(1.3, 1.3);
        self.alpha = 0.0;
    } completion:^(BOOL finished) {
        if (finished) {
            [self removeFromSuperview];
        }
    }];
}

#pragma mark - Instance Methods
- (void)showInView:(UIView *)aView animated:(BOOL)animated
{
    [aView addSubview:self];
    if (animated) {
        [self fadeIn];
    }
    if (needRefresh) {
        [MBProgressHUD showHUDAddedTo:self animated:YES];
        [[HKMDiscoverer agent] queryLeads];
    }
}

#pragma mark - Tableview datasource & delegates
- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section {
    if ([HKMDiscoverer agent].leads != nil) {
        int rows = [[HKMDiscoverer agent].leads count];
        if ([HKMDiscoverer agent].contactsLoaded)
            return rows;
        else
            return rows+1;
    } else {
        return 1;
    }
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
    int rows = [[HKMDiscoverer agent].leads count];
    UITableViewCell *cell = nil;
    
    if (indexPath.row < rows) {
        cell = [tableView dequeueReusableCellWithIdentifier:@"HKINviteViewCell"];
        if (cell ==  nil) {
            cell = [[HKMInviteViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:@"HKINviteViewCell"];
        }
        HKMLead *lead = (HKMLead *)[[HKMDiscoverer agent].leads objectAtIndex:indexPath.row];
        cell.textLabel.text = lead.name;
        cell.detailTextLabel.text = lead.osType;
        cell.imageView.image = (lead.image) ? lead.image : [UIImage imageNamed:@"contact.png"];
        
        if (lead.selected) {
            cell.accessoryType = UITableViewCellAccessoryCheckmark;
        } else {
            cell.accessoryType = UITableViewCellAccessoryNone;
        }
    } else if (indexPath.row == rows && ![HKMDiscoverer agent].contactsLoaded) {
        cell = [tableView dequeueReusableCellWithIdentifier:@"HKINviteViewAccessContactCell"];
        if (cell ==  nil) {
            cell = [[HKMInviteViewAccessContactCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:@"HKINviteViewAccessContactCell"];
        }
        cell.textLabel.text = @"Select from Contacts..";
    } else {
        NSLog(@"ERROR!!!! rows=%d indexPath=%d", rows, indexPath.row);
        cell = [tableView dequeueReusableCellWithIdentifier:@"HKINviteViewCell"];
        if (cell ==  nil) {
            cell = [[HKMInviteViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:@"HKINviteViewCell"];
        }
        HKMLead *lead = (HKMLead *)[[HKMDiscoverer agent].leads objectAtIndex:indexPath.row];
        cell.textLabel.text = lead.name;
        cell.detailTextLabel.text = lead.osType;
        cell.imageView.image = (lead.image) ? lead.image : [UIImage imageNamed:@"contact.png"];
        
        if (lead.selected) {
            cell.accessoryType = UITableViewCellAccessoryCheckmark;
        } else {
            cell.accessoryType = UITableViewCellAccessoryNone;
        }
    }
    
    return cell;
}

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath {
    // this means contacts not loaded into leads
    if (indexPath.row == [HKMDiscoverer agent].leads.count) {
        // load address book contacts into the list
        [self launchWithPermissionCheckDiscover:NO];
        [[HKMDiscoverer agent] loadAddressBooktoIntoLeads];
    } else {
        HKMLead *lead = [[HKMDiscoverer agent].leads objectAtIndex:indexPath.row];
        if (lead.recordId == 0) {
            lead.selected = lead.selected ? NO : YES;
        } else {
            if (lead.selected)
                lead.selected = NO;
            else
                [[HKMDiscoverer agent] queryPhoneNumberTypeAtLeadsIndex:indexPath.row];
        }
    }
    
    [_tableView reloadData];
}

#pragma mark - TouchTouchTouch
- (void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event
{
    UITouch *touch = [touches anyObject];
    if ([touch view]  == self) {
        // tell the delegate the cancellation
        if (self.delegate && [self.delegate respondsToSelector:@selector(inviteCancelled)]) {
            [self.delegate inviteCancelled];
        }
        // dismiss self
        [self fadeOut];
    }
    
}

#pragma mark - DrawDrawDraw
- (void)drawRect:(CGRect)rect
{
    CGRect bgRect = CGRectInset(rect, POPLISTVIEW_SCREENINSET, POPLISTVIEW_SCREENINSET);
    CGRect titleRect = CGRectMake(POPLISTVIEW_SCREENINSET + 10, POPLISTVIEW_SCREENINSET + 10 + 5,
                                  rect.size.width -  2 * (POPLISTVIEW_SCREENINSET + 10), 30);
    CGRect separatorRect = CGRectMake(POPLISTVIEW_SCREENINSET, POPLISTVIEW_SCREENINSET + POPLISTVIEW_HEADER_HEIGHT - 2,
                                      rect.size.width - 2 * POPLISTVIEW_SCREENINSET, 2);
    
    CGContextRef ctx = UIGraphicsGetCurrentContext();
    
    // Draw the background with shadow
    CGContextSetShadowWithColor(ctx, CGSizeZero, 6., [UIColor colorWithWhite:0 alpha:.75].CGColor);
    [[UIColor colorWithWhite:0 alpha:.75] setFill];
    
    
    float x = POPLISTVIEW_SCREENINSET;
    float y = POPLISTVIEW_SCREENINSET;
    float width = bgRect.size.width;
    float height = bgRect.size.height;
    CGMutablePathRef path = CGPathCreateMutable();
	CGPathMoveToPoint(path, NULL, x, y + RADIUS);
	CGPathAddArcToPoint(path, NULL, x, y, x + RADIUS, y, RADIUS);
	CGPathAddArcToPoint(path, NULL, x + width, y, x + width, y + RADIUS, RADIUS);
	CGPathAddArcToPoint(path, NULL, x + width, y + height, x + width - RADIUS, y + height, RADIUS);
	CGPathAddArcToPoint(path, NULL, x, y + height, x, y + height - RADIUS, RADIUS);
	CGPathCloseSubpath(path);
	CGContextAddPath(ctx, path);
    CGContextFillPath(ctx);
    CGPathRelease(path);
    
    // Draw the title and the separator with shadow
    CGContextSetShadowWithColor(ctx, CGSizeMake(0, 1), 0.5f, [UIColor blackColor].CGColor);
    [[UIColor colorWithRed:0.020 green:0.549 blue:0.961 alpha:1.] setFill];
    [_title drawInRect:titleRect withFont:[UIFont systemFontOfSize:20.]];
    CGContextFillRect(ctx, separatorRect);
}

#pragma mark - UI interaction
- (void)sendInviteAction {
    _lastInviteCount = 0;
    NSMutableArray *phones = [NSMutableArray arrayWithCapacity:16];
    for (HKMLead *lead in [HKMDiscoverer agent].leads) {
        if (lead.selected) {
            [phones addObject:lead.phone];
            _lastInviteCount++;
        }
    }
    if ([phones count] > 0) {
        [MBProgressHUD showHUDAddedTo:self animated:YES];
        [[HKMDiscoverer agent] newReferral:phones useVirtualNumber:_useVirtualNumber];
    } else {
        MBProgressHUD *hud = [MBProgressHUD showHUDAddedTo:self animated:YES];
        hud.mode = MBProgressHUDModeText;
        hud.labelText = INVITE_SELECT_SOME_FRIENDS_MSG;
        dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, 2.0 * NSEC_PER_SEC);
        dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
            [MBProgressHUD hideHUDForView:self animated:YES];
        });
    }
}

#pragma mark - AGE address book handling functions
- (void) launchWithPermissionCheckDiscover: (BOOL)discover
{
    [MBProgressHUD hideHUDForView:self animated:YES];

    ABAddressBookRef ab = ABAddressBookCreate();
    if (ABAddressBookRequestAccessWithCompletion != NULL) {
        ABAddressBookRequestAccessWithCompletion(ab, ^(bool granted, CFErrorRef error) {
            if (granted) {
                if (discover) {
                    dispatch_async(dispatch_get_main_queue(), ^{
                        [MBProgressHUD showHUDAddedTo:self animated:YES];
                        int contactsToUpLoad = ([[HKMDiscoverer agent] lastDiscoverDate] == nil) ? INITIAL_ADDRESSBOOK_UPLOAD_SIZE : MAX_ADDRESSBOOK_UPLOAD_SIZE;
                        //NSLog(@"launchWithPermissionCheckDiscover - uploading %d contacts", contactsToUpLoad);
                        [[HKMDiscoverer agent] discover:contactsToUpLoad];
                    });
                }
            } else {
                dispatch_async(dispatch_get_main_queue(), ^{
                    UIAlertView* alert = [[UIAlertView alloc] initWithTitle:ADDRESSBOOK_DENIED_ACCESS_ALERT_TITLE message:ADDRESSBOOK_DENIED_ACCESS_ALERT_MSG delegate:self cancelButtonTitle:@"Dismiss" otherButtonTitles:nil];
                    [alert show];
                });
            }
        });
    } else {
        if (discover) {
            // iOS 5
            [MBProgressHUD showHUDAddedTo:self animated:YES];
            int contactsToUpLoad = ([[HKMDiscoverer agent] lastDiscoverDate] == nil) ? 200 : 2000;
            //NSLog(@"launchWithPermissionCheckDiscover - uploading %d contacts", contactsToUpLoad);
            [[HKMDiscoverer agent] discover:contactsToUpLoad];
        }
    }
}


// Handle alert view action asking user to enable access to address book
- (void)alertView:(UIAlertView *)actionSheet clickedButtonAtIndex:(NSInteger)buttonIndex {
    // the user clicked one of the OK/Cancel buttons
    if (buttonIndex == 0) {
        [self fadeOut];
    }
}

// register for notification of AGE related callback events
- (void)registerNotification
{
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(discoverCompleted) name:NOTIF_HOOK_DISCOVER_COMPLETE object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(discoverCompleted) name:NOTIF_HOOK_DISCOVER_NO_CHANGE object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(discoverFailed) name:NOTIF_HOOK_DISCOVER_FAILED object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(queryLeadsCompleted) name:NOTIF_HOOK_QUERY_ORDER_COMPLETE object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(queryLeadsFailed) name:NOTIF_HOOK_QUERY_ORDER_FAILED object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(networkError) name:NOTIF_HOOK_NETWORK_ERROR object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(newReferralCompleted) name:NOTIF_HOOK_NEW_REFERRAL_COMPLETE object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(queryPhoneNumberTypeCompleted) name:NOTIF_HOOK_QUERY_PHONE_NUMBER_TYPE_COMPLETE object:nil];

}

- (void) discoverCompleted
{
    if ([[HKMDiscoverer agent] lastDiscoverDate] == nil) {
        _aTimer = [NSTimer scheduledTimerWithTimeInterval:FIRST_USE_WAIT_TIME
                                                   target:self
                                                 selector:@selector(timerFired:)
                                                 userInfo:nil
                                                  repeats:NO];
    } else {
        [[HKMDiscoverer agent] queryLeads];
    }
}

- (void) discoverFailed {
    [MBProgressHUD hideHUDForView:self animated:YES];
    //[self fadeOut];
}

-(void)timerFired:(NSTimer *) theTimer
{
    NSLog(@"timerFired @ %@", [theTimer fireDate]);
    BOOL status = [[HKMDiscoverer agent] queryLeads];
    NSLog(@"timerFired - queryLeads status=%d", status);
}

- (void) queryLeadsCompleted
{
    needRefresh = false;
    [_tableView reloadData];
    [MBProgressHUD hideHUDForView:self animated:YES];
    
    if (firstTimeQueryLeadsCompleted) {
        MBProgressHUD *hud = [MBProgressHUD showHUDAddedTo:self animated:YES];
        hud.mode = MBProgressHUDModeText;
        hud.labelText = @"Pull down to refresh";
        dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, 1.5 * NSEC_PER_SEC);
        dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
            [MBProgressHUD hideHUDForView:self animated:YES];
        });
        firstTimeQueryLeadsCompleted = false;
    }

    int rowsOfData = [_tableView numberOfRowsInSection:0];
    if (rowsOfData == 1 && ![HKMDiscoverer agent].contactsLoaded) {
        [[HKMDiscoverer agent] loadAddressBooktoIntoLeads];
        [_tableView reloadData];
    }
}

- (void) queryLeadsFailed
{
    [MBProgressHUD hideHUDForView:self animated:YES];
    //[self fadeOut];
}

- (void) newReferralCompleted
{
    if (_useVirtualNumber) {
        needRefresh = true;
        [MBProgressHUD hideHUDForView:self animated:YES];
        MBProgressHUD *hud = [MBProgressHUD showHUDAddedTo:self animated:YES];
        hud.mode = MBProgressHUDModeText;
        hud.labelText = @"Invitation Sent!";
        dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, 1.0 * NSEC_PER_SEC);
        dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
            [MBProgressHUD hideHUDForView:self animated:YES];
            // tell the delegate the selection
            if (self.delegate && [self.delegate respondsToSelector:@selector(invitedCount:)]) {
                [self.delegate invitedCount:_lastInviteCount];
            }
            
            [self fadeOut];
        });
    } else {
        [MBProgressHUD hideHUDForView:self animated:YES];
        //        NSMutableArray *phones = [[NSMutableArray arrayWithCapacity:16] retain];
        NSMutableArray *phones = [NSMutableArray arrayWithCapacity:16];
        for (HKMLead *lead in [HKMDiscoverer agent].leads) {
            if (lead.selected) {
                [phones addObject:lead.phone];
            }
        }
        
        // replace any occurrance of template variable %invitee% with Hi
        [HKMDiscoverer agent].referralMessage = [[HKMDiscoverer agent].referralMessage stringByReplacingOccurrencesOfString:@"%invitee%" withString:@"Hi"];
        MFMessageComposeViewController *controller = [[MFMessageComposeViewController alloc] init];
        controller.body = [HKMDiscoverer agent].referralMessage;
        controller.recipients = phones;
        controller.messageComposeDelegate = self;
        [[self viewController] presentModalViewController:controller animated:YES];
    }
}

- (void) queryPhoneNumberTypeCompleted
{
    [_tableView reloadData];
    HKMLead *lead = [[HKMDiscoverer agent].leads objectAtIndex:[HKMDiscoverer agent].queryPhoneNumberTypeListIndex];
    if (!lead.selected) {
        // only create the toast if there isn't any existing toast.
        // Workaround for freezing toast.
        NSArray *array = [MBProgressHUD allHUDsForView:self];
        if ([array count] == 0) {
            MBProgressHUD *hud = [MBProgressHUD showHUDAddedTo:self animated:YES];
            hud.mode = MBProgressHUDModeText;
            hud.labelText = @"Not a mobile contact";
            dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, 1.0 * NSEC_PER_SEC);
            dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
                [MBProgressHUD hideHUDForView:self animated:YES];
            });
        }
    }
}

- (void) sendReferralViaVirtualNumberCompleted
{
}

- (void)messageComposeViewController:(MFMessageComposeViewController *)controller didFinishWithResult:(MessageComposeResult)result {
    
    [[self viewController] dismissModalViewControllerAnimated:YES];
    
    [MBProgressHUD hideHUDForView:self animated:YES];

    if (result == MessageComposeResultCancelled) {
        [[HKMDiscoverer agent] updateReferral:NO];
        _lastInviteCount = 0;
        if (self.delegate && [self.delegate respondsToSelector:@selector(inviteCancelled)]) {
            [self.delegate inviteCancelled];
        }
    } else {
        [[HKMDiscoverer agent] updateReferral:YES];
        if (self.delegate && [self.delegate respondsToSelector:@selector(invitedCount:)]) {
            [self.delegate invitedCount:_lastInviteCount];
        }

        MBProgressHUD *hud = [MBProgressHUD showHUDAddedTo:self animated:YES];
        hud.mode = MBProgressHUDModeText;
        hud.labelText = @"Invitation Sent!";
        dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, 1.0 * NSEC_PER_SEC);
        dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
            [MBProgressHUD hideHUDForView:self animated:YES];
            [self fadeOut];
        });
        needRefresh = true;
    }
}

- (UIViewController *)viewController
{
    for (UIResponder * nextResponder = self.nextResponder;
         nextResponder;
         nextResponder = nextResponder.nextResponder)
    {
        if ([nextResponder isKindOfClass:[UIViewController class]])
            return (UIViewController *)nextResponder;
    }
    
    // Not found
    NSLog(@"%@ doesn't seem to have a viewController", self);
    return nil;
}

- (void) networkError
{
    NSLog(@"networkError invoked");
    [MBProgressHUD hideHUDForView:self animated:YES];
    
    UIAlertView* alert = [[UIAlertView alloc] init];
	alert.title = NO_NETWORK_COVERAGE_ERROR_ALERT_TITLE;
	alert.message = NO_NETWORK_COVERAGE_ERROR_ALERT_MSG;
	[alert addButtonWithTitle:@"OK"];
	alert.cancelButtonIndex = 0;
	[alert show];
    
    [self fadeOut];
}

@end
