# Introduction

AGE Invitation UI is a powerful drop-in IOS controls that you can deploy to your app within minutes.  AGE Invitation UI will help your users identify and share your app with friends & associates having compatible mobile devices and highest LTV.  For complete feature set and how it works, please visit <a href="http://www.hookmobile.com"  target="_blank">Hook Mobile</a>.

# Getting Sample App Up and Running
A sample app is included with this project that demonstrate AGE Invitation UI launched from a button.  Once you download and open the Xcode project, click on Run button in Xcode.  You may launch the sample app in the simulator or your IOS device.  We recommend you to run the sample app on your iPhone so you can feel the magic of how easy it is to invite your contacts list from your address book.  

[![](https://dl.dropbox.com/s/izxzj9qxrgl2axd/AGEUI1.PNG)](https://www.dropbox.com/s/izxzj9qxrgl2axd/AGEUI1.PNG)
[![](https://dl.dropbox.com/s/pm1uzrjn1p1dk9v/AGEUI2.PNG)](https://www.dropbox.com/s/pm1uzrjn1p1dk9v/AGEUI2.PNG)

When AGE Invitation UI is invoked for first time, it will analyze the address book.  It may take a few seconds before the list of suggested contacts is displayed.  You may use swipe-pull gesture to force a refresh of the list.  The list of contacts shown in the list is filtered by criteria you define for your app profile in our developer portal. 

Select one or more entries from the suggested and click on the <b>Invite</b> button to fire off the invitation text message.  The recipient(s) of the invitation will receive a personalized text message on their phone:

[![](https://dl.dropbox.com/s/zg3qbf5ac8om7cg/inviteSms.PNG)](https://dl.dropbox.com/s/zg3qbf5ac8om7cg/inviteSms.PNG)

The message is completely customizable by you, and it can be further personalized to include the sender and app name.

# Integration Setup
Now you understand how easy it is for users to use AGE and how powerful it can be for you to market your app. Let’s start integrate AGE into your app. Please follow the following steps. 

1. <b><a href="http://www.hookmobile.com"  target="_blank">sign up</a></b> for an API Key for your app at Hook Mobile developer portal. This API Key will be used in the next section when you start doing your integration. 

2. Copy all the necessary files into your Xcode project.  In your Finder, locate and drag <b>AGE</b> sub-folder into your Xcode Project.  AGE Invitation UI requires linking with following system libraries:

	* AddressBook.Framework
	* MessageUI.Framework
	* QuartzCore.Framework
	* CoreTelephony.Framework
	* CoreGraphics.Framework

AGE Invitation UI depends on following third party open-source libraries.  Source codes for these libraries are already included in this project, so <b>no additional</b> files need to be copied into your Xcode project.

* <a href="https://github.com/jdg/MBProgressHUD" target="_blank">MBProgressHUD</a>
* <a href="https://github.com/Sephiroth87/ODRefreshControl" target="_blank">ODRefreshControl</a>
* <a href="https://github.com/ylechelle/OpenUDID" target="_blank">OpenUDID</a>
* <a href="https://github.com/groopd/SBJSON-library" target="_blank">SBJson</a>

# ARC 
AGE Invitation UI is written to work with ARC enabled project.  

# Integration Point

AGE Invitation UI should literally take minutes to integrate because there is only two  integration points to your existing app.  The first integration point is to inform AGE when your app initially loads.  The following code snippet shows the newly added code in <b>BOLD</b>.  Be sure to replace the <code>AGE_API_KEY</code> with key assigned to your application during registration.

<pre><code>
#import "AppDelegate.h"
<b>#import "HKMInviteView.h"</b>

..

- (BOOL)application:(UIApplication *)application didFinishLaunchingWithOptions:(NSDictionary *)launchOptions
{
    <b>[HKMInviteView activateWithAppKey:AGE_API_KEY];</b>
..
}
</code></pre>

The second integration point is where you want to invoke the user invitation flow.  You will need to decide when and how to trigger the AGE Invitation UI.  <i>E.g. Do you want to trigger it from a Share button or from an time or conditional event?</i> 

The following sample code snipet declares a method <code>inviteClicked</code> for launching AGE Invitation UI.  In this method, we create an <code>HKMInviteView</code> object, aka AGE Invitation UI.  The <code>initWithTitle:sendBtnLabel:view:</code> constructor takes 3 arguments.  The first argument is the title of the AGE Invitiation UI.  The second argument is the button label.  The third argument is the parent view of the AGE Invitation UI. The fourth argument determine if SMS invitation will be sent from user's phone number or from Virtual Number.  The last argument is the name of the inviter.

<pre><code>- (IBAction) inviteClicked:(id)sender
{
    if (inviteView == nil) {
        inviteView = [[HKMInviteView alloc] initWithTitle:@"Suggested Contacts"
                                             sendBtnLabel:@"Invite"
                                                     view:self.view
                                         useVirtualNumber:NO];
        inviteView.delegate = self;
    }
    [inviteView showInView:self.view animated:YES];
}</code></pre>

Note that in above code snippet, self is set as the delegate for receiving message from <code>inviteView</code>.  The assigned delegate can receive two types of messages:
<code>invitedCount:count</code>
<code>invitedCancelled</code> 

<code>invitedCount:count</code> message is sent to delegate when app user invited one or more recipients so the host app can keep counter of invitation sent.  The <code>invitedCancelled</code> message is sent to delegate when app user cancelled out of the AGE Invitation UI dialog.  In order to receive these two messages, your app must implement the delegate protocol <code>HKMInviteDelegate</code>

Below sample code snippet illustrates implementation of the delegate protocol by the host app:

<pre><code>#pragma mark - HKInvite delegates
- (void)invitedCount:(NSInteger)count;
{
    _infoLabel.text = [NSString stringWithFormat:@"You have shared this app with %d friends", count];
}

- (void)inviteCancelled
{
    _infoLabel.text = @"You have cancelled from sharing";
}
</code></pre>


#UI Customization
You may customize the look and feel of the AGE Invitation UI specifically for your app.  There are few things you can easily tweak with minimal effort:

* Dimension of AGE Invitation UI - You can change the value of <code>#define POPLISTVIEW_SCREENINSET</code> constant defined in <code>HKInviteView.m</code>
* Default contact image - You can replace the default image contact.png with another image you provide.
* Landscape View - The AGE Invitation UI supports both portrait and landscape view with no modification needed.

If you want to build UI by yourself while still using our AGE features. You can achieve your goal by implementing our core <a href="https://github.com/hookmobile/App-Growth-Engine-iOS-SDK" target="_blank">AGE SDK</a>.
